from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

from asciimatics.event import KeyboardEvent, MouseEvent  # type: ignore
from asciimatics.exceptions import NextScene, StopApplication  # type: ignore
from asciimatics.screen import Screen  # type: ignore
from asciimatics.widgets import Frame  # type: ignore

if TYPE_CHECKING:
    from hesiod.ui import TUI


class BaseForm(ABC, Frame):
    EDIT_FORM = "EDIT"
    RECAP_FORM = "RECAP"
    LAST_ROW = -3

    def __init__(
        self,
        name: str,
        screen: Screen,
        parent: "TUI",
        previous_form: Optional[str] = None,
        next_form: Optional[str] = None,
    ) -> None:
        """Base class for TUI forms.

        Args:
            name: The name of the form.
            screen: The screen where the form will be displayed.
            parent: The parent TUI.
            previous: The name of the previous form (optional).
            next: The name of the next form (optional).
        """
        ABC.__init__(self)
        Frame.__init__(
            self,
            screen,
            int(screen.height),
            int(screen.width),
            name=name,
            on_load=self.refresh,
        )
        self.set_theme("bright")
        self.parent = parent
        self.previous_form = previous_form
        self.next_form = next_form
        self.draw()

    @abstractmethod
    def draw(self) -> None:
        """Draw the form, adding the needed widgets."""
        pass

    def refresh(self) -> None:
        """Refresh the form."""
        pass

    def process_event(self, event: Union[MouseEvent, KeyboardEvent]) -> None:
        """Process either a mouse or a keyboard event.

        Args:
            event : The event that triggered the function.
        """
        if isinstance(event, KeyboardEvent):
            if event.key_code == Screen.ctrl("n"):
                self.next()
            elif event.key_code == Screen.ctrl("b") and self.previous_form is not None:
                self.back()
        Frame.process_event(self, event)

    @abstractmethod
    def before_exit(self) -> None:
        """Perform final operations when exiting the form."""
        pass

    def next(self) -> None:
        """Exit this form and move to the next one."""
        self.before_exit()
        if self.next_form is not None:
            raise NextScene(self.next_form)
        else:
            raise StopApplication("Stop")

    def back(self) -> None:
        """Move to the previous form."""
        if self.previous_form is not None:
            raise NextScene(self.previous_form)



from typing import Optional

from asciimatics.event import Event, KeyboardEvent, MouseEvent  # type: ignore
from asciimatics.screen import Screen  # type: ignore
from asciimatics.widgets import DatePicker, _DatePickerPopup  # type: ignore


class CustomDatePickerPopup(_DatePickerPopup):
    def _on_close(self, cancelled: bool) -> None:
        """When closing the popup, in any case, give the
        focus back to the parent widget.

        Args:
            cancelled: a flag that indicates if the selection was cancelled.
        """
        _DatePickerPopup._on_close(self, cancelled)
        self._parent.focus()


class CustomDatePicker(DatePicker):
    def process_event(self, event: Optional[Event]) -> Optional[Event]:
        """Process either a keyboard or a mouse event. If the user pressed
        enter/space or double clicked on the widget, a popup will be shown
        to allow the user to pick a date.

        Args:
            event: the event to be handled.

        Returns:
            The handled event, in case somebody else needs it.
        """
        if event is not None:
            if isinstance(event, KeyboardEvent):
                if event.key_code in [Screen.ctrl("M"), Screen.ctrl("J"), ord(" ")]:
                    event = None
            elif isinstance(event, MouseEvent):
                if event.buttons != 0:
                    if self.is_mouse_over(event, include_label=False):
                        event = None
            if event is None:
                self._child = CustomDatePickerPopup(self, year_range=self._year_range)
                self.frame.scene.add_effect(self._child)

        return event



from typing import TYPE_CHECKING

from asciimatics.parsers import AsciimaticsParser  # type: ignore
from asciimatics.screen import Screen  # type: ignore
from asciimatics.widgets import Divider, Layout, Text, TextBox  # type: ignore

from hesiod.cfg.cfghandler import RUN_NAME_KEY
from hesiod.ui.tui.baseform import BaseForm
from hesiod.ui.tui.widgets.wgtfactory import WidgetFactory

if TYPE_CHECKING:
    from hesiod.ui import TUI


class RecapForm(BaseForm):
    TITLE = "Recap (^B: back - ^N: save)"
    RUN_NAME = "RUN NAME:"

    def __init__(self, screen: Screen, parent_app: "TUI") -> None:
        """Create new form that show a recap of the run config.

        Args:
            screen: The screen where the form will be displayed.
            parent_app: The parent TUI.
        """
        BaseForm.__init__(
            self,
            BaseForm.RECAP_FORM,
            screen,
            parent_app,
            previous_form=BaseForm.EDIT_FORM,
        )
        self.title = RecapForm.TITLE
        self.palette["disabled"] = self.palette["edit_text"]

    def draw(self) -> None:
        layout = Layout([100])
        self.add_layout(layout)

        self.recap_text_box = TextBox(1, parser=AsciimaticsParser())
        self.recap_text_box.disabled = True
        layout.add_widget(self.recap_text_box)

        layout.add_widget(Divider())

        self.run_name_widget = Text(name=RUN_NAME_KEY, label=RecapForm.RUN_NAME)
        self.run_name_widget.value = ""
        layout.add_widget(self.run_name_widget)

        self.fix()

    def refresh(self) -> None:
        label_style = (self.palette["label"][0], self.palette["label"][1])
        text_style = (self.palette["edit_text"][0], self.palette["edit_text"][1])
        recap = WidgetFactory.get_recap_text(self.parent.run_cfg, label_style, text_style)
        self.recap_text_box._required_height = len(recap)
        self.recap_text_box.value = recap

        self.fix()

    def before_exit(self) -> None:
        """Save run name in the parent app config when exiting the form."""
        run_name = self.run_name_widget.value
        self.parent.run_cfg[RUN_NAME_KEY] = run_name



import pkg_resources

from hesiod.core import get_cfg_copy, get_out_dir, get_run_name, hcfg, hmain

__version__ = pkg_resources.get_distribution("hesiod").version
__all__ = [
    "__version__",
    "hmain",
    "hcfg",
    "get_cfg_copy",
    "get_out_dir",
    "get_run_name",
    "parse_args",
]



import re
from abc import ABC, abstractmethod
from ast import literal_eval
from datetime import date, datetime
from pathlib import Path
from typing import Any, List, Optional, Tuple, Type

from asciimatics.widgets import Label, Text, Widget  # type: ignore

from hesiod.cfg.cfghandler import CFG_T
from hesiod.ui.tui.widgets.custom.datepicker import CustomDatePicker
from hesiod.ui.tui.widgets.custom.dropdown import CustomDropdownList
from hesiod.ui.tui.widgets.custom.filebrowser import CustomFileBrowser
from hesiod.ui.tui.widgets.custom.radiobuttons import CustomRadioButtons
from hesiod.ui.tui.widgets.wgthandler import BaseWidgetHandler, BoolWidgetHandler
from hesiod.ui.tui.widgets.wgthandler import OptionsWidgetHandler, WidgetHandler

WGT_T = Tuple[Optional[WidgetHandler], Widget]


class WidgetParser(ABC):
    PREFIX = "   "
    DATE_PATTERN = r"^@DATE$"
    DEFAULT_DATE_PATTERN = r"^@DATE\((today|Today|TODAY|\d{4}-\d{2}-\d{2})\)$"
    FILE_PATTERN = r"^@FILE$"
    DEFAULT_FILE_PATTERN = r"^@FILE\(.+\)$"
    BASE_PATTERN = r"^@BASE\([0-9A-Za-z_.]+\)$"
    OPTIONS_PATTERN = r"^@OPTIONS\((.+)\)$"
    BOOL_PATTERN = r"^@BOOL\((true|True|TRUE|false|False|FALSE)\)$"

    @staticmethod
    def match(s: str, pattern: str) -> bool:
        """Check if a given string matches a pattern.

        Args:
            s: The string to be checked.
            pattern: The pattern to check.

        Returns:
            True if the given string matches the patter, False otherwise.
        """
        return bool(re.match(pattern, s))

    @staticmethod
    @abstractmethod
    def can_handle(x: Any) -> bool:
        """Check if the input config can be handled by this parser.

        Args:
            x: The input config.

        Returns:
            True if the input config is int, float, str or list.
        """
        ...

    @staticmethod
    @abstractmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        """Parse a literal config and return a list with the needed widgets.

        Args:
            cfg_key: The name of the config.
            label_prefix: The prefix for the name of the config.
            cfg_value: The literal value of the config.
            base_cfg_dir: The path to the base configs directory.

        Returns:
            A list with the widgets for the given config.
        """
        ...


class LiteralWidgetParser(WidgetParser):
    @staticmethod
    def can_handle(x: Any) -> bool:
        return type(x) in [int, float, str, bool, list, tuple, set, date]

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        handler = WidgetHandler(cfg_key)

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label}:"

        widget = Text(label=label, name=cfg_key)
        widget.value = str(cfg_value)

        return [(handler, widget)]


class DateWidgetParser(WidgetParser):
    TODAY = "today"
    FORMAT = r"%Y-%M-%d"
    HINT = "(↲ to select)"

    @staticmethod
    def can_handle(x: Any) -> bool:
        if isinstance(x, str):
            can_handle = WidgetParser.match(x, WidgetParser.DATE_PATTERN)
            can_handle = can_handle or WidgetParser.match(x, WidgetParser.DEFAULT_DATE_PATTERN)
            return can_handle
        return False

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        handler = WidgetHandler(cfg_key)

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label} {DateWidgetParser.HINT}:"

        widget = CustomDatePicker(label=label, name=cfg_key)

        if WidgetParser.match(cfg_value, WidgetParser.DEFAULT_DATE_PATTERN):
            default = cfg_value.split("(")[-1].split(")")[0]
            default = default.lower()

            value = None
            if default == DateWidgetParser.TODAY:
                value = date.today()
            else:
                value = datetime.strptime(default, DateWidgetParser.FORMAT).date()

            if value is not None:
                widget.value = value

        return [(handler, widget)]


class FileWidgetParser(WidgetParser):
    HINT = "(↲ to select)"

    @staticmethod
    def can_handle(x: Any) -> bool:
        if isinstance(x, str):
            can_handle = WidgetParser.match(x, WidgetParser.FILE_PATTERN)
            can_handle = can_handle or WidgetParser.match(x, WidgetParser.DEFAULT_FILE_PATTERN)
            return can_handle
        return False

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        handler = WidgetHandler(cfg_key)

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label} {FileWidgetParser.HINT}:"

        if WidgetParser.match(cfg_value, WidgetParser.DEFAULT_FILE_PATTERN):
            default = cfg_value.split("(")[-1].split(")")[0]
            path = default.lower()
        else:
            path = str(Path(".").absolute())

        widget = CustomFileBrowser(label, cfg_key, path)

        return [(handler, widget)]


class BoolWidgetParser(WidgetParser):
    @staticmethod
    def can_handle(x: Any) -> bool:
        return isinstance(x, str) and WidgetParser.match(x, WidgetParser.BOOL_PATTERN)

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        handler = BoolWidgetHandler(cfg_key)

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label}:"

        values = [(BoolWidgetHandler.TRUE, 0), (BoolWidgetHandler.FALSE, 1)]
        widget = CustomRadioButtons(values, label=label, name=cfg_key)
        default = cfg_value.split("(")[-1].split(")")[0]
        default = str(default).lower()
        widget.value = 0 if default == BoolWidgetHandler.TRUE else 1

        return [(handler, widget)]


class OptionsWidgetParser(WidgetParser):
    @staticmethod
    def can_handle(x: Any) -> bool:
        return isinstance(x, str) and WidgetParser.match(x, WidgetParser.OPTIONS_PATTERN)

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        opt_search = re.search(WidgetParser.OPTIONS_PATTERN, cfg_value)
        if opt_search is not None:
            options = opt_search.group(1)
        else:
            raise ValueError("Something went wrong, no options passed to OptionsWidgetParser.")

        values = [(option.strip(), i) for i, option in enumerate(options.split(";"))]

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label}:"

        widget = CustomRadioButtons(values, label=label, name=cfg_key)

        handler_values: List[Any] = []
        for v, _ in values:
            try:
                value = literal_eval(v)
            except ValueError:
                value = v
            handler_values.append(value)
        handler = OptionsWidgetHandler(cfg_key, handler_values)

        return [(handler, widget)]


class BaseWidgetParser(WidgetParser):
    HINT = "(↲ to select)"

    @staticmethod
    def can_handle(x: Any) -> bool:
        return isinstance(x, str) and WidgetParser.match(x, WidgetParser.BASE_PATTERN)

    @staticmethod
    def get_files_list(dir: Path) -> List[Path]:
        """Returns the list of files contained in the given directory.

        Files are searched in the given directory and in all its subdirectories.

        Args:
            dir: The directory to be searched.

        Returns:
            The list of files.
        """
        files: List[Path] = []
        children = [p for p in dir.glob("*")]
        for p in children:
            if p.is_file():
                files.append(p)
            elif p.is_dir():
                files.extend(BaseWidgetParser.get_files_list(p))
        return files

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        base_key = cfg_value.split("(")[-1].split(")")[0]
        base_keys = base_key.split(".")

        root = base_cfg_dir
        for k in base_keys:
            subdirs = [d for d in root.glob(k) if d.is_dir()]
            if len(subdirs) != 1:
                raise ValueError(f"Cannot find base key {base_key}")
            root = subdirs[0]

        files = sorted(BaseWidgetParser.get_files_list(root))
        values = [(f.stem, i) for i, f in enumerate(files)]
        if len(values) == 0:
            raise ValueError(f"Cannot find any option for the base key {base_key}")

        base_keys = [str(f.relative_to(base_cfg_dir)) for f in files]
        base_keys = [k.split(".")[0] for k in base_keys]
        base_keys = [k.replace("/", ".") for k in base_keys]

        handler = BaseWidgetHandler(cfg_key, base_keys)

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label} {BaseWidgetParser.HINT}:"

        widget = CustomDropdownList(values, label=label, name=cfg_key)

        return [(handler, widget)]


class RecursiveWidgetParser(WidgetParser):
    @staticmethod
    def can_handle(x: Any) -> bool:
        return isinstance(x, dict)

    @staticmethod
    def parse(cfg_key: str, label_prefix: str, cfg_value: Any, base_cfg_dir: Path) -> List[WGT_T]:
        widgets: List[WGT_T] = []

        label = cfg_key.split(".")[-1]
        label = f"{label_prefix}{label}:"
        widgets.append((None, Label(label)))

        children_prefix = f"{label_prefix}{WidgetParser.PREFIX}"
        children = WidgetFactory.get_widgets(
            cfg_value, base_cfg_dir, cfg_prefix=cfg_key, label_prefix=children_prefix
        )
        widgets.extend(children)

        return widgets


class WidgetFactory:
    @staticmethod
    def get_parsers() -> List[Type[WidgetParser]]:
        """Create a list of widget parsers.

        Important: parsers are checked in the same order in which they
        are inserted into the list.

        Returns:
            The list of widget parsers.
        """
        parsers: List[Type[WidgetParser]] = []

        # specials
        parsers.append(BoolWidgetParser)
        parsers.append(DateWidgetParser)
        parsers.append(FileWidgetParser)
        parsers.append(OptionsWidgetParser)
        parsers.append(BaseWidgetParser)

        # recursive
        parsers.append(RecursiveWidgetParser)

        # literal
        parsers.append(LiteralWidgetParser)

        return parsers

    @staticmethod
    def get_widgets(
        cfg: CFG_T,
        base_cfg_dir: Path,
        cfg_prefix: str = "",
        label_prefix: str = "",
    ) -> List[WGT_T]:
        """Prepare widgets for a given config.

        Args:
            cfg: The config.
            base_cfg_dir: The path to the base configs directory.
            cfg_prefix: The prefix for the config key of the returned widgets.
            label_prefix: The prefix for the name of the returned widgets.

        Returns:
            The list of the widgets for the given config.
        """
        widgets: List[WGT_T] = []
        parsers = WidgetFactory.get_parsers()

        for k in cfg:
            for parser in parsers:
                if parser.can_handle(cfg[k]):
                    cfg_key = f"{cfg_prefix}.{k}" if len(cfg_prefix) > 0 else k
                    widgets.extend(parser.parse(cfg_key, label_prefix, cfg[k], base_cfg_dir))
                    break

        return widgets

    @staticmethod
    def get_recap_text(
        cfg: CFG_T,
        label_style: Tuple[int, int],
        text_style: Tuple[int, int],
    ) -> List[str]:
        """Create a recap text for a given config.

        Args:
            cfg: The config to recap.

        Returns:
            A list with the lines of the recap text.
        """
        recap: List[str] = []

        lbl = "${" + str(label_style[0]) + "," + str(label_style[1]) + "}"
        txt = "${" + str(text_style[0]) + "," + str(text_style[1]) + "}"

        for _, widget in WidgetFactory.get_widgets(cfg, Path()):
            label = widget.text if isinstance(widget, Label) else widget.label
            line = f"{lbl}{label}"
            if not isinstance(widget, Label):
                line += f"{txt} {widget.value}"
            recap.append(line)

        return recap



from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Type

from hesiod.cfg.cfgparser import CFG_T, ConfigParser
from hesiod.cfg.yamlparser import YAMLConfigParser

BASE_KEY = "base"
RUN_NAME_KEY = "run_name"


class ConfigHandler:
    @staticmethod
    def get_parser(ext: str) -> Type["ConfigParser"]:
        """Return the proper parser given the extension of the file to parse.

        Args:
            ext: The extension of the file to parse.

        Returns:
            The proper parser for the given extension.
        """
        parsers = [YAMLConfigParser]
        parsers_table: Dict[str, Type[ConfigParser]] = {}
        for parser in parsers:
            for managed_ext in parser.get_managed_extensions():
                parsers_table[managed_ext] = parser

        ext = ext.strip()
        ext = ext[1:] if ext[0] == "." else ext
        return parsers_table[ext]

    @staticmethod
    def load_cfg(run_cfg_file: Path, base_cfg_dir: Path) -> CFG_T:
        """Load config replacing "bases" with proper values.

        Args:
            run_cfg_file: The path to the run config file.
            base_cfg_dir: The path to the base configs directory.

        Returns:
            The loaded config.
        """

        cfg = ConfigHandler.load_cfg_file(run_cfg_file)

        base_cfgs = ConfigHandler.load_base_cfgs(base_cfg_dir)

        cfg = ConfigHandler.replace_bases(cfg, base_cfgs)

        return cfg

    @staticmethod
    def load_cfg_file(cfg_file: Path) -> CFG_T:
        """Load config from a given file.

        Args:
            cfg_file: The file with the config.

        Raises:
            ValueError: If the config is not of the expected type.

        Returns:
            The loaded config.
        """
        parser = ConfigHandler.get_parser(cfg_file.suffix)
        cfg = parser.read_cfg_file(cfg_file)

        if not isinstance(cfg, dict):
            raise ValueError(f"Error in {cfg_file.name}: Config should be a dictionary.")

        for key in cfg:
            if not isinstance(key, str):
                raise ValueError("Error in {cfg_file.name}: Config keys should be strings.")

        return cfg

    @staticmethod
    def load_cfg_dir(cfg_dir: Path) -> Dict[str, CFG_T]:
        """Load configs recursively from a given directory.

        Args:
            cfg_dir: The config directory.

        Returns:
            The loaded config.
        """
        cfg: Dict[str, Dict[str, Any]] = {}

        cfg_files = [p for p in cfg_dir.glob("*.yaml")]
        for cfg_file in cfg_files:
            cfg[cfg_file.stem] = ConfigHandler.load_cfg_file(cfg_file)

        cfg_subdirs = [p for p in cfg_dir.glob("*") if p.is_dir()]
        for cfg_subdir in cfg_subdirs:
            cfg[cfg_subdir.name] = ConfigHandler.load_cfg_dir(cfg_subdir)

        return cfg

    @staticmethod
    def load_base_cfgs(base_cfg_dir: Path) -> Dict[str, CFG_T]:
        """Load all the base configs.

        Starting from the given path, each subdir is considered a
        base config. Then, each base config is loaded by reading
        recursively subdirs and files. Finally all base configs
        are saved in a global dictionary.

        Args:
            base_cfg_dir: The path with the base config directories.

        Returns:
            A dictionary with all the base configs.
        """
        base_cfgs: Dict[str, CFG_T] = {}
        cfg_files = [p for p in base_cfg_dir.glob("*.yaml")]
        for cfg_file in cfg_files:
            base_cfgs[cfg_file.stem] = ConfigHandler.load_cfg_file(cfg_file)
        cfg_dirs = [p for p in base_cfg_dir.glob("*") if p.is_dir()]
        for cfg_dir in cfg_dirs:
            base_cfgs[cfg_dir.name] = ConfigHandler.load_cfg_dir(cfg_dir)

        return base_cfgs

    @staticmethod
    def replace_base(
        cfg: CFG_T,
        base_cfgs: Dict[str, CFG_T],
        base_key: str = BASE_KEY,
    ) -> CFG_T:
        """Replace base placeholder in a given config.

        Args:
            cfg: The config with base placeholder.
            base_cfgs: The available base configs.
            base_key: The string used as base key.

        Raises:
            ValueError: If it is not possible to retrieve the base config.

        Returns:
            The config with the base placeholder replaced.
        """
        new_cfg = deepcopy(cfg)
        base_id = new_cfg[base_key]
        base_cfg = base_cfgs

        for k in base_id.split("."):
            if k not in base_cfg:
                raise ValueError(f"Config error: cannot find base {base_id}")
            base_cfg = base_cfg[k]

        del new_cfg[base_key]

        for k in base_cfg:
            if k not in new_cfg:
                new_cfg[k] = base_cfg[k]

        return new_cfg

    @classmethod
    def replace_bases(
        cls,
        cfg: CFG_T,
        base_cfgs: Dict[str, CFG_T],
        base_key: str = BASE_KEY,
    ) -> CFG_T:
        """Replace all the bases in a given config recursively.

        Args:
            cfg: The config to process.
            base_cfgs: The available base configs.
            base_key: The string used as base key.

        Returns:
            The config with all bases resolved.
        """
        new_cfg = deepcopy(cfg)

        # replace base in main cfg
        while base_key in new_cfg:
            new_cfg = cls.replace_base(new_cfg, base_cfgs, base_key=base_key)

        # replace bases in sub cfgs
        for cfg_key in new_cfg:
            if isinstance(new_cfg[cfg_key], dict):
                new_cfg[cfg_key] = cls.replace_bases(new_cfg[cfg_key], base_cfgs, base_key=base_key)

        return new_cfg

    @staticmethod
    def save_cfg(cfg: CFG_T, cfg_file: Path) -> None:
        """Save config into the given file.

        Args:
            cfg: The config to be saved.
            cfg_file: The path to the output file.
        """
        parser = ConfigHandler.get_parser(cfg_file.suffix)
        parser.write_cfg(cfg, cfg_file)



from abc import ABC, abstractmethod
from pathlib import Path

from hesiod.cfg.cfghandler import CFG_T


class UI(ABC):
    def __init__(
        self,
        template_cfg: CFG_T,
        base_cfg_dir: Path,
    ) -> None:
        """Create a new user interface (UI).

        Args:
            template_file: The path to the config template file.
            base_cfg_dir: The path to the base configs directory.
        """
        ABC.__init__(self)
        self.template_cfg = template_cfg
        self.base_cfg_dir = base_cfg_dir

    @abstractmethod
    def show(self) -> CFG_T:
        ...



from typing import TYPE_CHECKING, List, Optional, Tuple

from asciimatics.screen import Screen  # type: ignore
from asciimatics.widgets import Layout, Widget  # type: ignore

from hesiod.cfg.cfghandler import CFG_T, ConfigHandler
from hesiod.ui.tui.baseform import BaseForm
from hesiod.ui.tui.widgets.wgtfactory import WidgetFactory
from hesiod.ui.tui.widgets.wgthandler import BaseWidgetHandler, WidgetHandler

if TYPE_CHECKING:
    from hesiod.ui import TUI


class EditForm(BaseForm):
    TITLE = "Edit configuration (^N: next)"

    def __init__(self, screen: Screen, parent: "TUI") -> None:
        """Create a new form that allows editing the run configuration.

        Args:
            screen: The screen where the form will be displayed.
            parent: The parent TUI.
        """
        BaseForm.__init__(self, BaseForm.EDIT_FORM, screen, parent, next_form=BaseForm.RECAP_FORM)
        self.title = EditForm.TITLE

    def draw(self) -> None:
        template_cfg = self.parent.template_cfg
        base_cfg_dir = self.parent.base_cfg_dir

        layout = Layout([100])
        self.add_layout(layout)

        self.widgets: List[Tuple[Optional[WidgetHandler], Widget]] = []
        for handler, widget in WidgetFactory.get_widgets(template_cfg, base_cfg_dir):
            layout.add_widget(widget)
            self.widgets.append((handler, widget))

        self.fix()

    def before_exit(self) -> None:
        """Save the config edited by the user.

        The config is saved in the parent app when exiting the form.
        """
        edited_cfg: CFG_T = {}

        for handler, widget in self.widgets:
            if handler is not None:
                edited_cfg = handler.update_cfg(edited_cfg, widget)

        base_cfgs = ConfigHandler.load_base_cfgs(self.parent.base_cfg_dir)
        temp_cfg = ConfigHandler.replace_bases(
            edited_cfg, base_cfgs, base_key=BaseWidgetHandler.BASE_KEY
        )
        run_cfg = ConfigHandler.replace_bases(temp_cfg, base_cfgs)

        self.parent.run_cfg.update(run_cfg)



from ast import literal_eval
from copy import deepcopy
from typing import Any, List

from asciimatics.widgets import Widget  # type: ignore

from hesiod.cfg.cfghandler import CFG_T


class WidgetHandler:
    def __init__(self, cfg_key: str) -> None:
        """Create a handler for a widget.

        Args:
            cfg_key: The key of the handled config.
        """
        self.cfg_key = cfg_key

    def get_value(self, widget: Widget) -> Any:
        """Extract the value from the given widget.

        Args:
            widget: The widget with the value of interest.

        Returns:
            The value extracted from the given widget.
        """
        try:
            value = literal_eval(widget.value)
        except (ValueError, SyntaxError):
            value = widget.value

        return value

    def update_cfg(self, cfg: CFG_T, widget: Widget) -> CFG_T:
        """Update the given config with the widget value.

        The given config is updated by setting the value extracted
        from the given widget in the right place.

        Args:
            cfg: The config to be updated.
            widget: The widget with the value of interest.

        Returns:
            The updated config.
        """
        updated_cfg = deepcopy(cfg)

        keys = self.cfg_key.split(".")
        curr_cfg = updated_cfg
        for key in keys[:-1]:
            if key not in curr_cfg:
                curr_cfg[key] = {}
            curr_cfg = curr_cfg[key]

        value = self.get_value(widget)
        curr_cfg[keys[-1]] = value

        return updated_cfg


class BoolWidgetHandler(WidgetHandler):
    TRUE = "true"
    FALSE = "false"

    def __init__(self, cfg_key: str) -> None:
        WidgetHandler.__init__(self, cfg_key)

    def get_value(self, widget: Widget) -> Any:
        selected_value = [BoolWidgetHandler.TRUE, BoolWidgetHandler.FALSE][widget.value]
        return selected_value == BoolWidgetHandler.TRUE


class OptionsWidgetHandler(WidgetHandler):
    def __init__(self, cfg_key: str, options: List[Any]) -> None:
        WidgetHandler.__init__(self, cfg_key)
        self.options = options

    def get_value(self, widget: Widget) -> Any:
        selected_value = widget.value
        return self.options[selected_value]


class BaseWidgetHandler(WidgetHandler):
    BASE_KEY = "***TUI_BASE***"

    def __init__(self, cfg_key: str, options: List[str]) -> None:
        WidgetHandler.__init__(self, cfg_key)
        self.options = options

    def get_value(self, widget: Widget) -> Any:
        selected_value = widget.value
        return {BaseWidgetHandler.BASE_KEY: self.options[selected_value]}



import functools
import re
import sys
from ast import literal_eval
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, List, Optional, Type, TypeVar, Union, cast

from typeguard import check_type

from hesiod.cfg.cfghandler import CFG_T, RUN_NAME_KEY, ConfigHandler
from hesiod.ui import TUI

T = TypeVar("T")
FUNCTION_T = Callable[..., Any]
_CFG: CFG_T = {}
RUN_FILE_NAME = "run.yaml"
OUT_DIR_KEY = "***hesiod_out_dir***"
RUN_NAME_STRATEGY_DATE = "date"
RUN_NAME_DATE_FORMAT = "%Y-%m-%d-%H-%M-%S"


def _parse_args(args: List[str]) -> None:
    """Parse the given args and add them to the global config.

    Each arg is expected with the format "{prefix}{key}{sep}{value}".
    {prefix} is optional and can be any amount of the char "-".
    {sep} is mandatory and can be one of "=", ":".
    {key} cannot contain the chars "-", "=" and ":".
    {value} can contain everything.

    Args:
        args: The list of args to be parsed.

    Raises:
        ValueError: If one of the given args is a not supported format.
    """
    for arg in args:
        pattern = r"^-*(?P<key>[^-=:]+)[=:]{1}(?P<value>.+)$"
        match = re.match(pattern, arg)

        if match is None:
            raise ValueError(f"One of the arg is in a not supported format {arg}.")
        else:
            while arg[0] == "-":
                arg = arg[1:]

            key = match.group("key")
            value = match.group("value")

            try:
                value = literal_eval(value)
            except (ValueError, SyntaxError):
                pass

            key_splits = key.split(".")
            cfg = _CFG
            for key in key_splits[:-1]:
                if key not in cfg:
                    cfg[key] = {}
                cfg = cfg[key]

            last_key = key_splits[-1]
            if last_key not in cfg:
                cfg[last_key] = {}
            cfg[last_key] = value


def _get_cfg(
    base_cfg_path: Path,
    template_cfg_path: Optional[Path],
    run_cfg_path: Optional[Path],
) -> CFG_T:
    """Load config either from template file or from run file.

    Args:
        base_cfg_path: The path to the directory with all the config files.
        template_cfg_path: The path to the template config file for this run.
        run_cfg_path: The path to the config file created by the user for this run.

    Raises:
        ValueError: If both template_cfg_path and run_cfg_path are None.

    Returns:
        The loaded config.
    """
    if run_cfg_path is not None:
        return ConfigHandler.load_cfg(run_cfg_path, base_cfg_path)
    elif template_cfg_path is not None:
        template_cfg = ConfigHandler.load_cfg(template_cfg_path, base_cfg_path)
        tui = TUI(template_cfg, base_cfg_path)
        return tui.show()
    else:
        msg = "Either a valid run file or a template file must be passed to hesiod."
        raise ValueError(msg)


def _get_default_run_name(strategy: str) -> str:
    """Get a run name according the given strategy.

    Args:
        strategy: The strategy to use to create the run name.

    Returns:
        The created run name.
    """
    run_name = ""
    if strategy == RUN_NAME_STRATEGY_DATE:
        now = datetime.now()
        run_name = now.strftime(RUN_NAME_DATE_FORMAT)

    return run_name


def _create_out_dir_and_save_run_file(
    cfg: CFG_T,
    out_dir_root: str,
    run_cfg_path: Optional[Path],
) -> None:
    """Create output directory for the current run.

    A new directory is created for the current run
    and the run file is saved in it (if needed).

    Args:
        cfg: The loaded config.
        out_dir_root: The root for output directories.
        run_cfg_path: The path to the config file created by the user for this run.

    Raises:
        ValueError: If the run name is not specified in the given config.
    """
    run_name = cfg.get(RUN_NAME_KEY, "")
    if run_name == "":
        msg = f"The config must contain a valid name for the run (key={RUN_NAME_KEY})."
        raise ValueError(msg)

    run_dir = Path(out_dir_root) / Path(run_name)
    run_file = run_dir / RUN_FILE_NAME

    create_dir = True
    if run_cfg_path is not None:
        create_dir = run_file.absolute() != run_cfg_path.absolute()

    if create_dir:
        run_dir.mkdir(parents=True, exist_ok=False)
        cfg[OUT_DIR_KEY] = str(run_dir.absolute())
        ConfigHandler.save_cfg(cfg, run_file)


def hmain(
    base_cfg_dir: Union[str, Path],
    template_cfg_file: Optional[Union[str, Path]] = None,
    run_cfg_file: Optional[Union[str, Path]] = None,
    create_out_dir: bool = True,
    out_dir_root: str = "logs",
    run_name_strategy: Optional[str] = RUN_NAME_STRATEGY_DATE,
    parse_cmd_line: bool = True,
) -> Callable[[FUNCTION_T], FUNCTION_T]:
    """Decorator for a given function.

    The decorator loads the configuration with the right parser
    and runs the given function.

    Args:
        base_cfg_dir: The path to the directory with all the config files.
        template_cfg_file: The path to the template config file for this run.
        run_cfg_file: The path to the config file created by the user for this run.
        create_out_dir: A flag that indicates whether hesiod should create
            an output directory or not.
        out_dir_root: The root for output directories.
        run_name_strategy: The strategy to assign a default run name if this is
            not specified by user (available options: "date").
        parse_cmd_line: A flag that indicates whether hesiod should parse args
            from the command line or not.

    Raises:
        ValueError: If both template_cfg_file and run_cfg_file are None.
        ValueError: If hesiod is asked to parse the command line and one
            of the args is in a not supported format.
        ValueError: If the run name is not specified in the run file
            and no default strategy is specified.

    Returns:
        Function wrapped in the decorator.
    """

    def decorator(fn: FUNCTION_T) -> FUNCTION_T:
        @functools.wraps(fn)
        def decorated_fn(*args: Any, **kwargs: Any) -> Any:
            global _CFG

            bcfg_path = Path(base_cfg_dir)
            run_cfg_path = Path(run_cfg_file) if run_cfg_file else None
            template_cfg_path = Path(template_cfg_file) if template_cfg_file else None

            _CFG = _get_cfg(bcfg_path, template_cfg_path, run_cfg_path)

            if parse_cmd_line and len(sys.argv) > 1:
                _parse_args(sys.argv[1:])

            run_name = _CFG.get(RUN_NAME_KEY, "")
            if run_name == "" and run_name_strategy is not None:
                run_name = _get_default_run_name(run_name_strategy)
                _CFG[RUN_NAME_KEY] = run_name

            if run_name == "":
                msg = (
                    f"A valid name must be provided for the run. Provide one "
                    f"by setting a value for the key {RUN_NAME_KEY} or "
                    f'selecting a default strategy (e.g. "date")'
                )
                raise ValueError(msg)

            if create_out_dir:
                _create_out_dir_and_save_run_file(_CFG, out_dir_root, run_cfg_path)

            return fn(*args, **kwargs)

        return decorated_fn

    return decorator


def hcfg(name: str, t: Optional[Type[T]] = None) -> T:
    """Get the requested parameter from global configuration.

    Args:
        name: name of the parameter.
        t: type of the parameter.

    Raises:
        TypeError: if the requested parameter is not of the required type.

    Returns:
        The requested parameter.
    """
    value = _CFG
    for n in name.split("."):
        value = value[n]

    if t is not None:
        check_type(name, value, t)

    value = deepcopy(value)

    return cast(T, value)


def get_cfg_copy() -> CFG_T:
    """Return a copy of the global configuration.

    Returns:
        A copy of the global configuration.
    """
    return deepcopy(_CFG)


def get_out_dir() -> Path:
    """Get path to the output directory for the current run.

    Returns:
        Path to the output directory.
    """
    out_dir = deepcopy(_CFG[OUT_DIR_KEY])
    return Path(out_dir)


def get_run_name() -> str:
    """Get the name of the current run.

    Raises:
        ValueError: if the current run has no name (it should never happen).

    Returns:
        The name of the current run.
    """
    run_name = _CFG.get(RUN_NAME_KEY, "")
    if run_name == "":
        raise ValueError("Something went wrong: current run has no name.")

    return run_name



from pathlib import Path
from typing import Any, List, Tuple

from ruamel.yaml import YAML
from ruamel.yaml.constructor import SafeConstructor
from ruamel.yaml.representer import Representer, SafeRepresenter

from hesiod.cfg.cfgparser import CFG_T, ConfigParser


def construct_python_tuple(constructor: SafeConstructor, node: Any) -> Tuple:
    return tuple(constructor.construct_sequence(node))


SafeConstructor.add_constructor("tag:yaml.org,2002:python/tuple", construct_python_tuple)
SafeRepresenter.add_representer(tuple, Representer.represent_tuple)


class YAMLConfigParser(ConfigParser):
    @staticmethod
    def get_managed_extensions() -> List[str]:
        return ["yaml"]

    @staticmethod
    def read_cfg_file(cfg_file: Path) -> CFG_T:
        yaml = YAML(typ="safe")
        return yaml.load(cfg_file)

    @staticmethod
    def write_cfg(cfg: CFG_T, cfg_file: Path) -> None:
        yaml = YAML(typ="safe")
        yaml.dump(cfg, cfg_file)
